//
//     Copyright (C) 2006 - 2007 SophiaCradle, Inc.
//     All rights reserved.
//
//      AstronomicalMath.cpp
//

#include "AstronomicalMath.hpp"

Void AstronomicalMath::Date::Set(UInt16 year, UInt16 month, UInt16 day, UInt16 hour, UInt16 minute, UInt16 second)
{
    SInt32 yyyy(static_cast<SInt32>(year)), mm(static_cast<SInt32>(month)), dd(static_cast<SInt32>(day)), dayCount;

    // Days from 0001/01/01 are requested by using Fairfield expressions
    if (mm < 3) {
        --yyyy;
        mm += 12;
    }
    dayCount = 365 * yyyy + yyyy / 4 - yyyy / 100 + yyyy / 400 + 306 * (mm + 1) / 10 - 428 + dd;

    // Elapsed days from 2001/01/01 are requested
    _days = static_cast<Float64>(dayCount) - 730486.0;
    _days += static_cast<Float64>(hour) / 24.0;
    _days += static_cast<Float64>(minute) / 1440.0;
    _days += static_cast<Float64>(second) / 86400.0;

    _dateUpdated = false;
    _j2000Updated = false;
}

Void AstronomicalMath::Date::Set(SFXDateConstRef date)
{
    SFXDateConst dateOrigin(2001, 1, 1, 0, 0, 0);
    SFXDateDurationConst span(date - dateOrigin);

    _days = static_cast<Float64>(span.AsSInt64()) / 86400.0;
    _year = date.GetYear();
    _month = date.GetMonth();
    _day = date.GetDay();
    _hour = date.GetHour();
    _minute = date.GetMinute();
    _second = date.GetSecond();

    _dateUpdated = true;
    _j2000Updated = false;
}

Void AstronomicalMath::Date::Set(DateConstRef ob)
{
    _year = ob._year;
    _month = ob._month;
    _day = ob._day;
    _hour = ob._hour;
    _minute = ob._minute;
    _second = ob._second;
    _days = ob._days;
    _j2000 = ob._j2000;
    _dateUpdated = ob._dateUpdated;
    _j2000Updated = ob._j2000Updated;
}

AstronomicalMath::Date::Date(Void) : _year(0), _month(0), _day(0), _hour(0), _minute(0), _second(0), _days(0.0), _j2000(0.0), _dateUpdated(false), _j2000Updated(false)
{
}

AstronomicalMath::Date::Date(UInt16 year, UInt16 month, UInt16 day, UInt16 hour, UInt16 minute, UInt16 second) : _year(0), _month(0), _day(0), _hour(0), _minute(0), _second(0), _days(0.0), _j2000(0.0), _dateUpdated(false), _j2000Updated(false)
{
    Set(year, month, day, hour, minute, second);
}

AstronomicalMath::Date::Date(SFXDateConstRef date)
{
    Set(date);
}

AstronomicalMath::Date::Date(Float64 days) : _year(0), _month(0), _day(0), _hour(0), _minute(0), _second(0), _days(0.0), _j2000(0.0), _dateUpdated(false), _j2000Updated(false)
{
    Set(days);
}

AstronomicalMath::Date::Date(DateConstRef ob)
{
    Set(ob);
}

AstronomicalMath::Date::~Date(Void)
{
}

//
// Elapsed days from 2001/01/01 00:00:00 are rectified at the date.
//
Void AstronomicalMath::Date::CalcDate(Void) const
{
    SInt32 days, k, l, m, n, d, yyyy, mm, dd;
    Float64 t, u;

    if (_dateUpdated) {
        return;
    }

    t = floor(_days);
    days = static_cast<SInt32>(t) + 730486;     // The 0001/01/01 regarded as 1 to total days
    t = _days - t;

    d = days + 305;
    k = d / 146097;
    d -= k * 146097;
    l = d / 36524;
    d -= l * 36524;
    m = d / 1461;
    d -= m * 1461;
    n = d / 365;
    yyyy = 400 * k + 100 * l + 4 * m + n;
    d = days - (365 * (yyyy - 1) + yyyy / 4 - yyyy / 100 + yyyy / 400 + 60);
    if (d == -1) {
        --yyyy;
        d = 365;
    }
    for (mm = 3; mm < 15; ++mm) {
        if (d >= 306 * (mm + 1) / 10 - 122 && d < 306 * (mm + 2) / 10 - 122) {
            break;
        }
    }
    dd = d - (306 * (mm + 1) / 10 - 123);
    if (mm > 12) {
        ++yyyy;
        mm -= 12;
    }

    _year = static_cast<UInt16>(yyyy);
    _month = static_cast<UInt16>(mm);
    _day = static_cast<UInt16>(dd);

    t *= 24.0;
    u = floor(t);
    _hour = static_cast<UInt16>(u);
    t = (t - u) * 60.0;
    u = floor(t);
    _minute = static_cast<UInt16>(u);
    t = (t - u) * 60.0;
    u = floor(t + 0.5);
    _second = static_cast<UInt16>(u);

    _dateUpdated = true;
}

//
// The passage Julius year from J2000.0 (2000/01/01 00:00 (GMT)) is calculated.
//
Void AstronomicalMath::Date::CalcJ2000Year(Void) const
{
    SInt32 y, m, d;
    Float64 k, k2, dk, dt;

    if (_j2000Updated) {
        return;
    }
    CalcDate();

    d = _day;
    if (_month < 3) {
        y = _year - 2001;
        m = _month + 12;
    }
    else {
        y = _year - 2000;
        m = _month;
    }

    k2 = static_cast<Float64>(365 * y + 30 * m + d) - 33.5 + floor(static_cast<Float64>(3 * (m + 1)) / 5.0) + floor(static_cast<Float64>(y) / 4.0);
    dt = floor(0.707018 * static_cast<Float64>(_year) - 1349.54);       // Earth rotation correction
    dk = _days - floor(_days);
    k = k2 + dk + dt / 86400.0;
    _j2000 = k / 365.25;

    _j2000Updated = true;
}

AstronomicalMath::DateDuration::DateDuration(Void) : _duration(0.0)
{
}

AstronomicalMath::DateDuration::DateDuration(SInt32 days, UInt16 hour, UInt16 minute, UInt16 second)
{
    Set(days, hour, minute, second);
}

AstronomicalMath::DateDuration::DateDuration(Float64 days)
{
    Set(days);
}

AstronomicalMath::DateDuration::DateDuration(SFXDateDurationConstRef duration)
{
    Set(duration);
}

AstronomicalMath::DateDuration::DateDuration(DateDurationConstRef ob)
{
    Set(ob);
}

AstronomicalMath::DateDuration::~DateDuration(Void)
{
}

Void AstronomicalMath::Star::Set(DateConstRef gmtDate)
{
    _gmtDate = gmtDate;
    _eclipticPosition.Longitude = CalcEclipticLongitude(_gmtDate.GetJ2000Year());
    _eclipticPosition.Latitude = CalcEclipticLatitude(_gmtDate.GetJ2000Year());
    _PI = CalcParallax(_gmtDate.GetJ2000Year());
}

AstronomicalMath::Star::Star(Void) : _PI(0.0)
{
    _eclipticPosition.Longitude = 0.0;
    _eclipticPosition.Latitude = 0.0;
}

AstronomicalMath::Star::Star(DateConstRef gmtDate)
{
    Set(gmtDate);
}

//
// Equatorial coordinates of the stars are acquired.
//
AstronomicalMath::AstronomicalPoint AstronomicalMath::Star::GetEquatorialPosition(Void) const
{
    Float64Const beta(DegreeToRadian(_eclipticPosition.Latitude));
    Float64Const lambda(DegreeToRadian(_eclipticPosition.Longitude));
    Float64Const eps(DegreeToRadian(CalcObliquityOfTheEcliptic(_gmtDate.GetJ2000Year())));
    Float64 u, v, w, alpha, delta;
    AstronomicalPoint result;

    u = cos(beta) * cos(lambda);
    v = - sin(beta) * sin(eps) + cos(beta) * sin(lambda) * cos(eps);
    w = sin(beta) * cos(eps) + cos(beta) * sin(lambda) * sin(eps);
    alpha = atan2(v, u);
    delta = atan(w / sqrt(u * u + v * v));
    if (alpha < 0.0) {
        alpha += 2.0 * ASTRONOMICALMATH_PI;
    }

    result.Longitude = RadianToDegree(alpha);
    result.Latitude = RadianToDegree(delta);
    return result;
}

Void AstronomicalMath::Sun::Set(DateConstRef gmtDate)
{
    Float64 r;

    _gmtDate = gmtDate;
    _eclipticPosition.Longitude = CalcEclipticLongitude(_gmtDate.GetJ2000Year());
    _eclipticPosition.Latitude = 0.0;

    r = CalcDistance(_gmtDate.GetJ2000Year());
    _S = 0.266994 / r;
    _PI = 0.00244282 / r;
}

AstronomicalMath::Sun::Sun(Void) : _S(0.0)
{
}

AstronomicalMath::Sun::Sun(DateConstRef gmtDate)
{
    Set(gmtDate);
}

//
// The celestial longitude of the sun is calculated.
// T : Passage Julius year.
//
Float64 AstronomicalMath::Sun::CalcEclipticLongitude(Float64 T) const
{
    Float64 lambdaS;

    lambdaS = 280.4603 + 360.00769 * T
                + (1.9146 - 0.00005 * T) * sin(DegreeToRadian(357.538 + 359.991 * T))
                + 0.0200 * sin(DegreeToRadian(355.05 + 719.981 * T))
                + 0.0048 * sin(DegreeToRadian(234.95 + 19.341 * T))
                + 0.0020 * sin(DegreeToRadian(247.1 + 329.64 * T))
                + 0.0018 * sin(DegreeToRadian(297.8 + 4452.67 * T))
                + 0.0018 * sin(DegreeToRadian(251.3 + 0.20 * T))
                + 0.0015 * sin(DegreeToRadian(343.2 + 450.37 * T))
                + 0.0013 * sin(DegreeToRadian(81.4 * 225.18 * T))
                + 0.0008 * sin(DegreeToRadian(132.5 + 659.29 * T))
                + 0.0007 * sin(DegreeToRadian(153.3 + 90.38 * T))
                + 0.0007 * sin(DegreeToRadian(206.8 + 30.35 * T))
                + 0.0006 * sin(DegreeToRadian(29.8 + 337.18 * T))
                + 0.0005 * sin(DegreeToRadian(207.4 + 1.50 * T))
                + 0.0005 * sin(DegreeToRadian(291.2 + 22.81 * T))
                + 0.0004 * sin(DegreeToRadian(234.9 + 315.56 * T))
                + 0.0004 * sin(DegreeToRadian(157.3 + 299.30 * T))
                + 0.0004 * sin(DegreeToRadian(21.1 * 720.02 * T))
                + 0.0003 * sin(DegreeToRadian(352.5 + 1079.97 * T))
                + 0.0003 * sin(DegreeToRadian(329.7 + 44.43 * T));
    lambdaS = fmod(lambdaS, 360.0);
    return (lambdaS >= 0.0) ? lambdaS : (lambdaS + 360.0);
}

//
// The distance between the sun and the earth is requested.
// T : Passage Julius year.
//
Float64 AstronomicalMath::Sun::CalcDistance(Float64 T) const
{
    Float64 q;

    q = (0.007256 - 0.0000002 * T) * sin(DegreeToRadian(267.54 + 359.991 * T))
            + 0.000091 * sin(DegreeToRadian(265.1 + 719.98 * T))
            + 0.000030 * sin(DegreeToRadian(90.0))
            + 0.000013 * sin(DegreeToRadian(27.8 + 4452.67 * T))
            + 0.000007 * sin(DegreeToRadian(254.0 + 450.4 * T))
            + 0.000007 * sin(DegreeToRadian(156.0 + 329.6 * T));
    return pow(10.0, q);
}

AstronomicalMath::Moon::Moon(DateConstRef gmtDate)
{
    Set(gmtDate);
}

//
// The celestial longitude of the moon is calculated.
// T : Passage Julius year.
//
Float64 AstronomicalMath::Moon::CalcEclipticLongitude(Float64 T) const
{
    Float64 Am, lambda_m;

    Am = 0.0040 * sin(DegreeToRadian(119.5 + 1.33 * T))
        + 0.0020 * sin(DegreeToRadian(55.0 + 19.34 * T))
        + 0.0006 * sin(DegreeToRadian(71.0 + 0.2 * T))
        + 0.0006 * sin(DegreeToRadian(54.0 + 19.3 * T));

    lambda_m = 218.3161 + 4812.67881 * T
                + 6.2887 * sin(DegreeToRadian(134.961 + 4771.9886 * T + Am))
                + 1.2740 * sin(DegreeToRadian(100.738 + 4133.3536 * T))
                + 0.6583 * sin(DegreeToRadian(235.700 + 8905.3422 * T))
                + 0.2136 * sin(DegreeToRadian(269.926 + 9543.9773 * T))
                + 0.1856 * sin(DegreeToRadian(177.525 + 359.9905 * T))
                + 0.1143 * sin(DegreeToRadian(6.546 + 9664.0404 * T))
                + 0.0588 * sin(DegreeToRadian(214.22 + 638.635 * T))
                + 0.0572 * sin(DegreeToRadian(103.21 + 3773.363 * T))
                + 0.0533 * sin(DegreeToRadian(10.66 + 13677.331 * T))
                + 0.0459 * sin(DegreeToRadian(238.18 + 8545.352 * T))
                + 0.0410 * sin(DegreeToRadian(137.43 + 4411.998 * T))
                + 0.0348 * sin(DegreeToRadian(117.84 + 4452.671 * T))
                + 0.0305 * sin(DegreeToRadian(312.49 + 5131.979 * T))
                + 0.0153 * sin(DegreeToRadian(130.84 + 758.698 * T))
                + 0.0125 * sin(DegreeToRadian(141.51 + 14436.029 * T))
                + 0.0110 * sin(DegreeToRadian(231.59 + 4892.052 * T))
                + 0.0107 * sin(DegreeToRadian(336.44 + 13038.696 * T))
                + 0.0100 * sin(DegreeToRadian(44.89 + 14315.966 * T))
                + 0.0085 * sin(DegreeToRadian(201.5 + 8266.71 * T))
                + 0.0079 * sin(DegreeToRadian(278.2 + 4493.34 * T))
                + 0.0068 * sin(DegreeToRadian(53.2 + 9265.33 * T))
                + 0.0052 * sin(DegreeToRadian(197.2 + 319.32 * T))
                + 0.0050 * sin(DegreeToRadian(295.4 + 4812.66 * T))
                + 0.0048 * sin(DegreeToRadian(235.0 + 19.34 * T))
                + 0.0040 * sin(DegreeToRadian(13.2 + 13317.34 * T))
                + 0.0040 * sin(DegreeToRadian(145.6 + 18449.32 * T))
                + 0.0040 * sin(DegreeToRadian(119.5 + 1.33 * T))
                + 0.0039 * sin(DegreeToRadian(111.3 + 17810.68 * T))
                + 0.0037 * sin(DegreeToRadian(349.1 + 5410.62 * T))
                + 0.0027 * sin(DegreeToRadian(272.5 + 9183.99 * T))
                + 0.0026 * sin(DegreeToRadian(107.2 + 13797.39 * T))
                + 0.0024 * sin(DegreeToRadian(211.9 + 988.63 * T))
                + 0.0024 * sin(DegreeToRadian(252.8 + 9224.66 * T))
                + 0.0022 * sin(DegreeToRadian(240.6 + 8185.36 * T))
                + 0.0021 * sin(DegreeToRadian(87.5 + 9903.97 * T))
                + 0.0021 * sin(DegreeToRadian(175.1 + 719.98 * T))
                + 0.0021 * sin(DegreeToRadian(105.6 + 3413.37 * T))
                + 0.0020 * sin(DegreeToRadian(55.0 + 19.34 * T))
                + 0.0018 * sin(DegreeToRadian(4.1 + 4013.29 * T))
                + 0.0016 * sin(DegreeToRadian(242.2 + 18569.38 * T))
                + 0.0012 * sin(DegreeToRadian(339.0 + 12678.71 * T))
                + 0.0011 * sin(DegreeToRadian(276.5 + 19208.02 * T))
                + 0.0009 * sin(DegreeToRadian(218.0 + 8586.0 * T))
                + 0.0008 * sin(DegreeToRadian(188.0 + 14037.3 * T))
                + 0.0008 * sin(DegreeToRadian(204.0 + 7906.7 * T))
                + 0.0007 * sin(DegreeToRadian(140.0 + 4052.0 * T))
                + 0.0007 * sin(DegreeToRadian(275.0 + 4853.3 * T))
                + 0.0007 * sin(DegreeToRadian(216.0 + 278.6 * T))
                + 0.0006 * sin(DegreeToRadian(128.0 + 1118.7 * T))
                + 0.0005 * sin(DegreeToRadian(247.0 + 22582.7 * T))
                + 0.0005 * sin(DegreeToRadian(181.0 + 19088.0 * T))
                + 0.0005 * sin(DegreeToRadian(114.0 + 17450.7 * T))
                + 0.0005 * sin(DegreeToRadian(332.0 + 5091.3 * T))
                + 0.0004 * sin(DegreeToRadian(313.0 + 398.7 * T))
                + 0.0004 * sin(DegreeToRadian(278.0 + 120.1 * T))
                + 0.0004 * sin(DegreeToRadian(71.0 + 9584.7 * T))
                + 0.0004 * sin(DegreeToRadian(20.0 + 720.0 * T))
                + 0.0003 * sin(DegreeToRadian(83.0 + 3814.0 * T))
                + 0.0003 * sin(DegreeToRadian(66.0 + 3494.7 * T))
                + 0.0003 * sin(DegreeToRadian(147.0 + 18089.3 * T))
                + 0.0003 * sin(DegreeToRadian(311.0 + 5492.0 * T))
                + 0.0003 * sin(DegreeToRadian(161.0 + 40.7 * T))
                + 0.0003 * sin(DegreeToRadian(280.0 + 23221.3 * T));

    lambda_m = fmod(lambda_m, 360.0);
    return (lambda_m >= 0.0) ? lambda_m : (lambda_m + 360.0);
}

//
// The celestial latitude of the moon is calculated.
//
Float64 AstronomicalMath::Moon::CalcEclipticLatitude(Float64 T) const
{
    Float64 Bm, beta_m;

    Bm = 0.0267 * sin(DegreeToRadian(234.95 + 19.341 * T))
            + 0.0043 * sin(DegreeToRadian(322.1 + 19.36 * T))
            + 0.0040 * sin(DegreeToRadian(119.5 + 1.33 * T))
            + 0.0026 * sin(DegreeToRadian(55.0 + 19.34 * T))
            + 0.0005 * sin(DegreeToRadian(307.0 + 19.4 * T));

    beta_m = 5.1282 * sin(DegreeToRadian(93.273 + 4832.0202 * T + Bm))
                + 0.2806 * sin(DegreeToRadian(228.235 + 9604.0088 * T))
                + 0.2777 * sin(DegreeToRadian(138.311 + 60.0316 * T))
                + 0.1732 * sin(DegreeToRadian(142.427 + 4073.3220 * T))
                + 0.0554 * sin(DegreeToRadian(194.01 + 8965.374 * T))
                + 0.0463 * sin(DegreeToRadian(172.55 + 698.667 * T))
                + 0.0326 * sin(DegreeToRadian(328.96 + 13737.362 * T))
                + 0.0172 * sin(DegreeToRadian(3.18 + 14375.997 * T))
                + 0.0093 * sin(DegreeToRadian(277.4 + 8845.31 * T))
                + 0.0088 * sin(DegreeToRadian(176.7 + 4711.96 * T))
                + 0.0082 * sin(DegreeToRadian(144.9 + 3713.33 * T))
                + 0.0043 * sin(DegreeToRadian(307.6 + 5470.66 * T))
                + 0.0042 * sin(DegreeToRadian(103.9 + 18509.35 * T))
                + 0.0034 * sin(DegreeToRadian(319.9 + 4433.31 * T))
                + 0.0025 * sin(DegreeToRadian(196.5 + 8605.38 * T))
                + 0.0022 * sin(DegreeToRadian(331.4 + 13377.37 * T))
                + 0.0021 * sin(DegreeToRadian(170.1 + 1058.66 * T))
                + 0.0019 * sin(DegreeToRadian(230.7 + 9244.02 * T))
                + 0.0018 * sin(DegreeToRadian(243.3 + 8206.68 * T))
                + 0.0018 * sin(DegreeToRadian(270.8 + 5192.01 * T))
                + 0.0017 * sin(DegreeToRadian(99.8 + 14496.06 * T))
                + 0.0016 * sin(DegreeToRadian(135.7 + 420.02 * T))
                + 0.0015 * sin(DegreeToRadian(211.1 + 9284.69 * T))
                + 0.0015 * sin(DegreeToRadian(45.8 + 9964.00 * T))
                + 0.0014 * sin(DegreeToRadian(219.2 + 299.96 * T))
                + 0.0013 * sin(DegreeToRadian(95.8 + 4472.03 * T))
                + 0.0013 * sin(DegreeToRadian(155.4 + 379.35 * T))
                + 0.0012 * sin(DegreeToRadian(38.4 + 4812.68 * T))
                + 0.0012 * sin(DegreeToRadian(148.2 + 4851.36 * T))
                + 0.0011 * sin(DegreeToRadian(138.3 + 19147.99 * T))
                + 0.0010 * sin(DegreeToRadian(18.0 + 12978.66 * T))
                + 0.0008 * sin(DegreeToRadian(70.0 + 17870.7 * T))
                + 0.0008 * sin(DegreeToRadian(326.0 + 9724.1 * T))
                + 0.0007 * sin(DegreeToRadian(294.0 + 13098.7 * T))
                + 0.0006 * sin(DegreeToRadian(224.0 + 5590.7 * T))
                + 0.0006 * sin(DegreeToRadian(52.0 + 13617.3 * T))
                + 0.0005 * sin(DegreeToRadian(280.0 + 8485.3 * T))
                + 0.0005 * sin(DegreeToRadian(239.0 + 4193.4 * T))
                + 0.0004 * sin(DegreeToRadian(311.0 + 9483.9 * T))
                + 0.0004 * sin(DegreeToRadian(238.0 + 23281.3 * T))
                + 0.0004 * sin(DegreeToRadian(81.0 + 10242.6 * T))
                + 0.0004 * sin(DegreeToRadian(13.0 + 9325.4 * T))
                + 0.0004 * sin(DegreeToRadian(147.0 + 14097.4 * T))
                + 0.0003 * sin(DegreeToRadian(205.0 + 22642.7 * T))
                + 0.0003 * sin(DegreeToRadian(107.0 + 18149.4 * T))
                + 0.0003 * sin(DegreeToRadian(146.0 + 3353.3 * T))
                + 0.0003 * sin(DegreeToRadian(234.0 + 19268.0 * T));

    return beta_m;
}

//
// The parallax of the moon is calculated.
// T : Passage Julius year.
//
Float64 AstronomicalMath::Moon::CalcParallax(Float64 T) const
{
    Float64 p;

    p = 0.9507 * sin(DegreeToRadian(90.0))
            + 0.0518 * sin(DegreeToRadian(224.98 + 4771.989 * T))
            + 0.0095 * sin(DegreeToRadian(190.7 + 4133.35 * T))
            + 0.0078 * sin(DegreeToRadian(325.7 + 8905.34 * T))
            + 0.0028 * sin(DegreeToRadian(0.0 + 9543.98 * T))
            + 0.0009 * sin(DegreeToRadian(100.0 + 13677.3 * T))
            + 0.0005 * sin(DegreeToRadian(329.0 + 8545.4 * T))
            + 0.0004 * sin(DegreeToRadian(194.0 + 3773.4 * T))
            + 0.0003 * sin(DegreeToRadian(227.0 + 4412.0 * T));

    return p;
}

//
// The necessary conditions are set to calculate the movement time of the star.
// gmtDate : Initial time value. It calculates within the range of 24 hours,
// position : Observation point longitude and latitude,
// H : Observation point height (m),
// R : Atmospheric refraction.
//
Void AstronomicalMath::StarRiseSet::Set(DateConstRef gmtDate, AstronomicalPointConstRef position, Float64 H, Float64 R)
{
    _gmtDate = gmtDate;
    _E = 0.0353333 * sqrt(H);
    _R = R;
    _position = position;
    _riseDate.Clear();
    _setDate.Clear();
}

//
// The necessary conditions are set to calculate the movement time of the star.
// gmtDate : Initial time value. It calculates within the range of 24 hours,
// lambda : Observation point longitude,
// phi : Observation point latitude,
// position : Observation point longitude,
// H : Observation point height (m),
// R : Atmospheric refraction.
//
Void AstronomicalMath::StarRiseSet::Set(DateConstRef gmtDate, Float64 lambda, Float64 phi, Float64 H, Float64 R)
{
    AstronomicalPoint position;

    position.Longitude = lambda;
    position.Latitude = phi;
    Set(gmtDate, position, H, R);
}

AstronomicalMath::StarRiseSet::StarRiseSet(Void) : _E(0.0), _R(0.0)
{
    _position.Longitude = 0.0;
    _position.Latitude = 0.0;
}

//
// gmtDate : Initial time value. It calculates within the range of 24 hours,
// position : Observation point longitude,
// H : Observation point height (m),
// R : Atmospheric refraction.
//
AstronomicalMath::StarRiseSet::StarRiseSet(DateConstRef gmtDate, AstronomicalPointConstRef position, Float64 H, Float64 R)
{
    Set(gmtDate, position, H, R);
}

//
// gmtDate : Initial time value. It calculates within the range of 24 hours.
// lambda : Observation point longitude,
// phi : Observation point latitude,
// position : Observation point longitude,
// H : Hight of observation point (m),
// R : atmospheric refraction.
//
AstronomicalMath::StarRiseSet::StarRiseSet(DateConstRef gmtDate, Float64 lambda, Float64 phi, Float64 H, Float64 R)
{
    Set(gmtDate, lambda, phi, H, R);
}

//
// The specified movement time is calculated.
//
Void AstronomicalMath::StarRiseSet::Compute(Void)
{
    Date x[5], e[7];
    Float64 y[7];
    SInt32 i, e_count(0);

    _riseDate.Clear();
    _setDate.Clear();

    x[0] = _gmtDate;
    for (i = 1; i < 5; ++i) {
        x[i] = x[i - 1] + DateDuration(0, 6, 0, 0);
    }
    for (i = 0; i < 5; ++i) {
        y[i] = DiffEquation(x[i]);
    }

    e[e_count++] = x[0];
    for (i = 0; i < 4; ++i) {
        if (y[i + 1] == 0.0) {
            e[e_count++] = x[i + 1];
            continue;
        }
        if (DifferentSign(y[i], y[i + 1])) {
#if defined TARGET_COMPILER_ARMCPP
            e[e_count++] = SolveEquation(&StarRiseSet::DiffEquation, x[i], x[i + 1]);
#else
            e[e_count++] = SolveEquation(DiffEquation, x[i], x[i + 1]);
#endif
        }
    }
    if (e[e_count - 1] != x[4]) {
        e[e_count++] = x[4];
    }

    for (i = 0; i < e_count; ++i) {
        y[i] = Equation(e[i]);
    }
    --e_count;
    for (i = 0; i < e_count; ++i) {
        if (y[i] < 0.0 && y[i + 1] > 0.0) {
#if defined TARGET_COMPILER_ARMCPP
            _riseDate.Append(SolveEquation(&StarRiseSet::Equation, e[i], e[i + 1]));
#else
            _riseDate.Append(SolveEquation(Equation, e[i], e[i + 1]));
#endif
        }
        else if (y[i] > 0.0 && y[i + 1] < 0.0) {
#if defined TARGET_COMPILER_ARMCPP
            _setDate.Append(SolveEquation(&StarRiseSet::Equation, e[i], e[i + 1]));
#else
            _setDate.Append(SolveEquation(Equation, e[i], e[i + 1]));
#endif
        }
    }
}

//
// The given equation is solved using the dichotomy.
//
AstronomicalMath::Date AstronomicalMath::StarRiseSet::SolveEquation(PFUNC_EQUATION equation, DateConstRef initial, DateConstRef final) const
{
    Date x1(initial), x2, x3(final);
    DateDuration d1, d2;
    Float64 y1, y2, y3;

    y1 = (this->*equation)(initial);
    if (y1 == 0.0) {
        return initial;
    }
    y3 = (this->*equation)(final);
    if (y3 == 0.0) {
        return final;
    }
    if (!DifferentSign(y1, y3)) {
        return initial;
    }

    for ( ; ; ) {
        x2 = Date::GetMidpoint(x1, x3);
        d1 = x2 - x1;
        d2 = x3 - x2;
        if (d1.Get() < ASTRONOMICALMATH_EQ_EPS || d2.Get() < ASTRONOMICALMATH_EQ_EPS) {
            break;
        }
        y2 = (this->*equation)(x2);
        if (y2 == 0.0) {
            break;
        }
        if (DifferentSign(y1, y2)) {
            x3 = x2;
            y3 = y2;
        }
        else {
            x1 = x2;
            y1 = y2;
        }
    }

    return x2;
}

AstronomicalMath::SunRiseSunSet::SunRiseSunSet(Void)
{
}

AstronomicalMath::SunRiseSunSet::SunRiseSunSet(DateConstRef gmtDate, AstronomicalPointConstRef position, Float64 H, Float64 R) : StarRiseSet(gmtDate, position, H, R)
{
}

AstronomicalMath::SunRiseSunSet::SunRiseSunSet(DateConstRef gmtDate, Float64 lambda, Float64 phi, Float64 H, Float64 R) : StarRiseSet(gmtDate, lambda, phi, H, R)
{
}

//
// The sunset and sunrise equation.
//
Float64 AstronomicalMath::SunRiseSunSet::Equation(DateConstRef gmtDate) const
{
    Sun sun(gmtDate);
    AstronomicalPoint position;
    Float64 Theta, t, delta, phi, k;

    position = sun.GetEquatorialPosition();     // Equatorial coordinates of the sun
    Theta = GetSiderealTime(sun._gmtDate, _position.Longitude /* Observation point longitude*/);
    t = DegreeToRadian(Theta - position.Longitude);
    delta = DegreeToRadian(position.Latitude);
    phi = DegreeToRadian(_position.Latitude /* Observation point latitude*/);
    k = DegreeToRadian(- sun._S - _E - _R + sun._PI);

    return sin(delta) * sin(phi) + cos(delta) * cos(phi) * cos(t) - sin(k);
}

AstronomicalMath::MoonRiseMoonSet::MoonRiseMoonSet(Void)
{
}

AstronomicalMath::MoonRiseMoonSet::MoonRiseMoonSet(DateConstRef gmtDate, AstronomicalPointConstRef position, Float64 H, Float64 R) : StarRiseSet(gmtDate, position, H, R)
{
}

AstronomicalMath::MoonRiseMoonSet::MoonRiseMoonSet(DateConstRef gmtDate, Float64 lambda, Float64 phi, Float64 H, Float64 R) : StarRiseSet(gmtDate, lambda, phi, H, R)
{
}

//
// The moonset and moonrise equation.
//
Float64 AstronomicalMath::MoonRiseMoonSet::Equation(DateConstRef gmtDate) const
{
    Moon moon(gmtDate);
    AstronomicalPoint position;
    Float64 Theta, t, delta, phi, k;

    position = moon.GetEquatorialPosition();    // Equatorial coordinates of the moon
    Theta = GetSiderealTime(moon._gmtDate, _position.Longitude /* Observation point longitude */);
    t = DegreeToRadian(Theta - position.Longitude);
    delta = DegreeToRadian(position.Latitude);
    phi = DegreeToRadian(_position.Latitude /* Observation point latitude */);
    k = DegreeToRadian(- _E - _R + moon._PI);

    return sin(delta) * sin(phi) + cos(delta) * cos(phi) * cos(t) - sin(k);
}

//
// The sidereal time is acquired.
// lambda : Longitude.
//
Float64 AstronomicalMath::GetSiderealTime(DateConstRef gmtDate, Float64 lambda)
{
    Float64Const T(gmtDate.GetJ2000Year());
    Float64Const d(gmtDate.Get() - floor(gmtDate.Get()));
    Float64 theta;

    theta = 100.4606 + 360.007700536 * T + 0.00000003879 * T * T + 360.0 * d + lambda;
    theta = fmod(theta, 360.0);
    return (theta >= 0.0) ? theta : (theta + 360.0);
}

//
// Moon oldness is acquired.
//
Float64 AstronomicalMath::GetMoonAge(DateConstRef gmtDate)
{
    Sun sun(gmtDate);
    Moon moon(gmtDate);
    Date t0_date;
    Float64 d_lambda, t0, t;

    t0 = gmtDate.Get();
    t = t0;
    d_lambda = moon._eclipticPosition.Longitude - sun._eclipticPosition.Longitude;
    if (d_lambda < 0.0) {
        d_lambda += 360.0;
    }
    while (d_lambda >= ASTRONOMICALMATH_SUN_MOON_ANGLE_EPS) {
        t0 -= d_lambda / 12.1818;
        t0_date.Set(t0);
        sun.Set(t0_date);
        moon.Set(t0_date);
        d_lambda = moon._eclipticPosition.Longitude - sun._eclipticPosition.Longitude;
    }

    return t - t0;
}
