//
//     Copyright (C) 2006 - 2007 SophiaCradle, Inc.
//     All rights reserved.
//
//      AstronomicalMath.hpp
//

#ifndef __ASTRONOMICAL_MATH_HPP
#define __ASTRONOMICAL_MATH_HPP

#include <SophiaFramework.hpp>
#include "Array.hpp"

#define ASTRONOMICALMATH_DEFAULT_H          0.0         // Default value of height in the observation point (m)
#define ASTRONOMICALMATH_DEFAULT_R          0.585556    // Default atmospheric refraction
#define ASTRONOMICALMATH_PI                 3.141592653589793238462643383280
#define ASTRONOMICALMATH_EQ_EPS             1.0E-5      // Machine epsilon
#define ASTRONOMICALMATH_SUN_MOON_ANGLE_EPS 0.05

SFMTYPEDEFCLASS(AstronomicalMath)

class AstronomicalMath {

SFMSEALCONSTRUCT(AstronomicalMath)

public:

    SFMTYPEDEFCLASS(DateDuration)

    SFMTYPEDEFCLASS(Date)

    class Date {
    private:

        mutable UInt16 _year;
        mutable UInt16 _month;
        mutable UInt16 _day;
        mutable UInt16 _hour;
        mutable UInt16 _minute;
        mutable UInt16 _second;
        Float64 _days;                  // Elapsed days from 2001/01/01 00:00:00 (GMT)
        mutable Float64 _j2000;         // Elapsed Julius year from J2000.0
        mutable Bool _dateUpdated;      // If the date is updated or not
        mutable Bool _j2000Updated;     // If J2000.0 is updated or not

    public:

        Void Set(UInt16 year, UInt16 month, UInt16 day, UInt16 hour, UInt16 minute, UInt16 second);
        Void Set(SFXDateConstRef date);
        Void Set(Float64 days);
        Void Set(DateConstRef ob);
        explicit Date(Void);
        explicit Date(UInt16 year, UInt16 month, UInt16 day, UInt16 hour, UInt16 minute, UInt16 second);
        explicit Date(SFXDateConstRef date);
        explicit Date(Float64 days);
        Date(DateConstRef ob);
        virtual ~Date(Void);

        DateConstRef operator=(SFXDateConstRef date);
        DateConstRef operator=(DateConstRef ob);
        UInt16 GetYear(Void) const;
        UInt16 GetMonth(Void) const;
        UInt16 GetDay(Void) const;
        UInt16 GetHour(Void) const;
        UInt16 GetMinute(Void) const;
        UInt16 GetSecond(Void) const;
        Float64 Get(Void) const;
        Float64 GetJ2000Year(Void) const;
        SFXDate GetSFXDate(Void) const;
        static Date GetMidpoint(DateConstRef date1, DateConstRef date2);

        Bool operator==(SFXDateConstRef date) const;
        Bool operator==(DateConstRef date) const;
        Bool operator!=(SFXDateConstRef date) const;
        Bool operator!=(DateConstRef date) const;
        Bool operator>=(SFXDateConstRef date) const;
        Bool operator>=(DateConstRef date) const;
        Bool operator>(SFXDateConstRef date) const;
        Bool operator>(DateConstRef date) const;
        Bool operator<=(SFXDateConstRef date) const;
        Bool operator<=(DateConstRef date) const;
        Bool operator<(SFXDateConstRef date) const;
        Bool operator<(DateConstRef date) const;
        Date operator+(DateDurationConstRef duration) const;
        Date operator+(SFXDateDurationConstRef duration) const;
        Date operator-(DateDurationConstRef duration) const;
        Date operator-(SFXDateDurationConstRef durarion) const;
        DateDuration operator-(DateConstRef date) const;
        DateDuration operator-(SFXDateConstRef date) const;
        DateConstRef operator+=(DateDurationConstRef duration);
        DateConstRef operator+=(SFXDateDurationConstRef duration);
        DateConstRef operator-=(DateDurationConstRef duration);
        DateConstRef operator-=(SFXDateDurationConstRef duration);

    private:

        Void CalcDate(Void) const;
        Void CalcJ2000Year(Void) const;
    };

    // Class that handles days
    class DateDuration {
    private:

        Float64 _duration;

    public:

        Void Set(SInt32 days, UInt16 hour, UInt16 minute, UInt16 second);
        Void Set(Float64 days);
        Void Set(SFXDateDurationConstRef duration);
        Void Set(DateDurationConstRef ob);
        explicit DateDuration(Void);
        explicit DateDuration(SInt32 days, UInt16 hour, UInt16 minute, UInt16 second);
        explicit DateDuration(Float64 days);
        explicit DateDuration(SFXDateDurationConstRef duration);
        DateDuration(DateDurationConstRef ob);
        virtual ~DateDuration(Void);

        DateDurationConstRef operator=(SFXDateDurationConstRef duration);
        DateDurationConstRef operator=(DateDurationConstRef ob);
        Float64 Get(Void) const;
        SInt32 GetDays(Void) const;
        UInt16 GetHour(Void) const;
        UInt16 GetMinute(Void) const;
        UInt16 GetSecond(Void) const;
        SFXDateDuration GetSFXDateDuration(Void) const;

        Bool operator==(SFXDateDurationConstRef duration) const;
        Bool operator==(DateDurationConstRef duration) const;
        Bool operator!=(SFXDateDurationConstRef duration) const;
        Bool operator!=(DateDurationConstRef duration) const;
        Bool operator>=(SFXDateDurationConstRef duration) const;
        Bool operator>=(DateDurationConstRef duration) const;
        Bool operator>(SFXDateDurationConstRef duration) const;
        Bool operator>(DateDurationConstRef duration) const;
        Bool operator<=(SFXDateDurationConstRef duration) const;
        Bool operator<=(DateDurationConstRef duration) const;
        Bool operator<(SFXDateDurationConstRef duration) const;
        Bool operator<(DateDurationConstRef duration) const;
        Date operator+(DateConstRef date) const;
        Date operator+(SFXDateConstRef date) const;
        DateDuration operator+(DateDurationConstRef duration) const;
        DateDuration operator+(SFXDateDurationConstRef duration) const;
        DateDuration operator-(DateDurationConstRef duration) const;
        DateDuration operator-(SFXDateDurationConstRef duration) const;
        DateDurationConstRef operator+=(DateDurationConstRef duration);
        DateDurationConstRef operator+=(SFXDateDurationConstRef duration);
        DateDurationConstRef operator-=(DateDurationConstRef duration);
        DateDurationConstRef operator-=(SFXDateDurationConstRef duration);
    };

    // Position of celestial bodies in celestial coordinates
    SFMTYPEDEFSTRUCT(AstronomicalPoint)

    struct AstronomicalPoint {
        Float64 Longitude;
        Float64 Latitude;
    };

    SFMTYPEDEFCLASS(SunRiseSunSet)
    SFMTYPEDEFCLASS(MoonRiseMoonSet)

    // Class that maintains stars information
    SFMTYPEDEFCLASS(Star)

    class Star {

    private:

        friend class SunRiseSunSet;
        friend class MoonRiseMoonSet;
        friend class AstronomicalMath;

    protected:

        Date _gmtDate;                          // Date
        AstronomicalPoint _eclipticPosition;    // Position of star in the ecliptic coordinate system
        Float64 _PI;                            // Parallax

    public:

        virtual Void Set(DateConstRef gmtDate);
        explicit Star(Void);
        Star(DateConstRef gmtDate);
        DateConstRef GetGMTDate(Void) const;
        AstronomicalPointConstRef GetEclipticPosition(Void) const;
        Float64 GetParallax(Void) const;
        AstronomicalPoint GetEquatorialPosition(Void) const;

    protected:

        virtual Float64 CalcEclipticLongitude(Float64 T) const = 0;
        virtual Float64 CalcEclipticLatitude(Float64 T) const = 0;
        virtual Float64 CalcParallax(Float64 T) const = 0;
    };

    friend class Star;

    // Class that maintains information on the sun
    SFMTYPEDEFCLASS(Sun)

    class Sun : public Star {

    private:

        Float64 _S;     // Sun semidiameter
        friend class SunRiseSunSet;

    public:

        virtual Void Set(DateConstRef gmtDate);
        explicit Sun(Void);
        Sun(DateConstRef gmtDate);

    protected:

        virtual Float64 CalcEclipticLongitude(Float64 T) const;
        virtual Float64 CalcEclipticLatitude(Float64 /* T */) const;
        virtual Float64 CalcParallax(Float64 /* T */) const;
        Float64 CalcDistance(Float64 T) const;
    };

    friend class Sun;

    // Class that maintains information on the moon
    SFMTYPEDEFCLASS(Moon)

    class Moon : public Star {
    public:

        explicit Moon(DateConstRef gmtDate);

    protected:

        virtual Float64 CalcEclipticLongitude(Float64 T) const;
        virtual Float64 CalcEclipticLatitude(Float64 T) const;
        virtual Float64 CalcParallax(Float64 T) const;
    };

    friend class Moon;

    // Class that calculates star movements
    SFMTYPEDEFCLASS(StarRiseSet)

    class StarRiseSet {
    protected:

        Date _gmtDate;                  // Date (the movement time is calculated here within the range of 24 hours)
        Float64 _E;                     // The dip of horizon in observation point
        Float64 _R;                     // Atmospheric refraction
        AstronomicalPoint _position;    // Observation point longitude and latitude
        Array<Date> _riseDate;          // The sunrise date
        Array<Date> _setDate;           // The sunset date
        typedef Float64 (StarRiseSet::*PFUNC_EQUATION)(DateConstRef gmtTime) const;

    public:

        Void Set(DateConstRef gmtDate, AstronomicalPointConstRef position, Float64 H = ASTRONOMICALMATH_DEFAULT_H, Float64 R = ASTRONOMICALMATH_DEFAULT_R);
        Void Set(DateConstRef gmtDate, Float64 lambda, Float64 phi, Float64 H = ASTRONOMICALMATH_DEFAULT_H, Float64 R = ASTRONOMICALMATH_DEFAULT_R);
        explicit StarRiseSet(Void);
        explicit StarRiseSet(DateConstRef gmtDate, AstronomicalPointConstRef position, Float64 H = ASTRONOMICALMATH_DEFAULT_H, Float64 R = ASTRONOMICALMATH_DEFAULT_R);
        explicit StarRiseSet(DateConstRef gmtDate, Float64 lambda, Float64 phi, Float64 H = ASTRONOMICALMATH_DEFAULT_H, Float64 R = ASTRONOMICALMATH_DEFAULT_R);
        Void Compute(Void);
        SInt32 GetRiseDateSize(Void) const;
        DateConstRef GetRiseDate(SInt32 index) const;
        SInt32 GetSetDateSize(Void) const;
        DateConstRef GetSetDate(SInt32 index) const;

    protected:

        virtual Float64 Equation(DateConstRef gmtDate) const = 0;
        Float64 DiffEquation(DateConstRef gmtDate) const;
        Date SolveEquation(PFUNC_EQUATION equation, DateConstRef initial, DateConstRef final) const;
    };

    friend class StarRiseSet;

    // Class that calculates the sunrise and sunset times
    class SunRiseSunSet : public StarRiseSet {
    public:

        explicit SunRiseSunSet(Void);
        explicit SunRiseSunSet(DateConstRef gmtDate, AstronomicalPointConstRef position, Float64 H = ASTRONOMICALMATH_DEFAULT_H, Float64 R = ASTRONOMICALMATH_DEFAULT_R);
        explicit SunRiseSunSet(DateConstRef gmtDate, Float64 lambda, Float64 phi, Float64 H = ASTRONOMICALMATH_DEFAULT_H, Float64 R = ASTRONOMICALMATH_DEFAULT_R);

    protected:

        virtual Float64 Equation(DateConstRef gmtDate) const;
    };

    friend class SunRiseSunSet;

    // Class that calculates moonrise and moonset times
    class MoonRiseMoonSet : public StarRiseSet {
    public:

        explicit MoonRiseMoonSet(Void);
        explicit MoonRiseMoonSet(DateConstRef gmtDate, AstronomicalPointConstRef position, Float64 H = ASTRONOMICALMATH_DEFAULT_H, Float64 R = ASTRONOMICALMATH_DEFAULT_R);
        explicit MoonRiseMoonSet(DateConstRef gmtDate, Float64 lambda, Float64 phi, Float64 H = ASTRONOMICALMATH_DEFAULT_H, Float64 R = ASTRONOMICALMATH_DEFAULT_R);

    protected:

        virtual Float64 Equation(DateConstRef gmtDate) const;
    };

    friend class MoonRiseMoonSet;

private:

    static Float64 DegreeToRadian(Float64 x);
    static Float64 RadianToDegree(Float64 x);
    static Float64 CalcObliquityOfTheEcliptic(Float64 T);
    static Bool DifferentSign(Float64 a, Float64 b);

public:

    static Float64 GetSiderealTime(DateConstRef gmtDate, Float64 lambda);
    static Float64 GetMoonAge(DateConstRef gmtDate);
};

Bool operator==(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2);
Bool operator!=(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2);
Bool operator>=(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2);
Bool operator>(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2);
Bool operator<=(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2);
Bool operator<(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2);
Bool operator==(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2);
Bool operator!=(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2);
Bool operator>=(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2);
Bool operator>(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2);
Bool operator<=(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2);
Bool operator<(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2);
AstronomicalMath::Date operator+(SFXDateDurationConstRef duration, AstronomicalMath::DateConstRef date);
AstronomicalMath::DateDuration operator-(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2);
AstronomicalMath::Date operator+(SFXDateConstRef date, AstronomicalMath::DateDurationConstRef duration);
AstronomicalMath::DateDuration operator+(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2);
AstronomicalMath::DateDuration operator-(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2);

inline Float64 AstronomicalMath::DegreeToRadian(Float64 x)
{
    return (ASTRONOMICALMATH_PI / 180.0) * x;
}

inline Float64 AstronomicalMath::RadianToDegree(Float64 x)
{
    return (180.0 / ASTRONOMICALMATH_PI) * x;
}

inline Void AstronomicalMath::Date::Set(Float64 days)
{
    _days = days;
    _dateUpdated = false;
    _j2000Updated = false;
}

inline AstronomicalMath::DateConstRef AstronomicalMath::Date::operator=(SFXDateConstRef date)
{
    Set(date);
    return *this;
}

inline AstronomicalMath::DateConstRef AstronomicalMath::Date::operator=(DateConstRef ob)
{
    Set(ob);
    return *this;
}

inline UInt16 AstronomicalMath::Date::GetYear(Void) const
{
    CalcDate();
    return _year;
}

inline UInt16 AstronomicalMath::Date::GetMonth(Void) const
{
    CalcDate();
    return _month;
}

inline UInt16 AstronomicalMath::Date::GetDay(Void) const
{
    CalcDate();
    return _day;
}

inline UInt16 AstronomicalMath::Date::GetHour(Void) const
{
    CalcDate();
    return _hour;
}

inline UInt16 AstronomicalMath::Date::GetMinute(Void) const
{
    CalcDate();
    return _minute;
}

inline UInt16 AstronomicalMath::Date::GetSecond(Void) const
{
    CalcDate();
    return _second;
}

inline Float64 AstronomicalMath::Date::Get(Void) const
{
    return _days;
}

inline Float64 AstronomicalMath::Date::GetJ2000Year(Void) const
{
    CalcJ2000Year();
    return _j2000;
}

inline SFXDate AstronomicalMath::Date::GetSFXDate(Void) const
{
    CalcDate();
    return SFXDate(_year, _month, _day, _hour, _minute, _second);
}

inline AstronomicalMath::Date AstronomicalMath::Date::GetMidpoint(DateConstRef date1, DateConstRef date2)
{
    return Date((date1._days + date2._days) / 2.0);
}

inline Bool AstronomicalMath::Date::operator==(SFXDateConstRef date) const
{
    return operator==(Date(date));
}

inline Bool AstronomicalMath::Date::operator==(DateConstRef date) const
{
    return _days == date._days;
}

inline Bool operator==(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2)
{
    return date2.operator==(date1);
}

inline Bool AstronomicalMath::Date::operator!=(SFXDateConstRef date) const
{
    return operator!=(Date(date));
}

inline Bool AstronomicalMath::Date::operator!=(DateConstRef date) const
{
    return _days != date._days;
}

inline Bool operator!=(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2)
{
    return date2.operator!=(date1);
}

inline Bool AstronomicalMath::Date::operator>=(SFXDateConstRef date) const
{
    return operator>=(Date(date));
}

inline Bool AstronomicalMath::Date::operator>=(DateConstRef date) const
{
    return _days >= date._days;
}

inline Bool operator>=(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2)
{
    return date2.operator<=(date1);
}

inline Bool AstronomicalMath::Date::operator>(SFXDateConstRef date) const
{
    return operator>(Date(date));
}

inline Bool AstronomicalMath::Date::operator>(DateConstRef date) const
{
    return _days > date._days;
}

inline Bool operator>(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2)
{
    return date2.operator<(date1);
}

inline Bool AstronomicalMath::Date::operator<=(SFXDateConstRef date) const
{
    return operator<=(Date(date));
}

inline Bool AstronomicalMath::Date::operator<=(DateConstRef date) const
{
    return _days <= date._days;
}

inline Bool operator<=(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2)
{
    return date2.operator>=(date1);
}

inline Bool AstronomicalMath::Date::operator<(SFXDateConstRef date) const
{
    return operator<(Date(date));
}

inline Bool AstronomicalMath::Date::operator<(DateConstRef date) const
{
    return _days < date._days;
}

inline AstronomicalMath::Date AstronomicalMath::Date::operator+(DateDurationConstRef duration) const
{
    return Date(_days + duration.Get());
}

inline AstronomicalMath::Date AstronomicalMath::Date::operator+(SFXDateDurationConstRef duration) const
{
    return operator+(DateDuration(duration));
}

inline AstronomicalMath::Date AstronomicalMath::Date::operator-(DateDurationConstRef duration) const
{
    return Date(_days - duration.Get());
}

inline AstronomicalMath::Date AstronomicalMath::Date::operator-(SFXDateDurationConstRef duration) const
{
    return operator-(DateDuration(duration));
}

inline AstronomicalMath::DateDuration AstronomicalMath::Date::operator-(DateConstRef date) const
{
    return DateDuration(_days - date._days);
}

inline AstronomicalMath::DateDuration AstronomicalMath::Date::operator-(SFXDateConstRef date) const
{
    return operator-(Date(date));
}

inline AstronomicalMath::DateConstRef AstronomicalMath::Date::operator+=(DateDurationConstRef duration)
{
    _days += duration.Get();
    _dateUpdated = false;
    _j2000Updated = false;
    return *this;
}

inline AstronomicalMath::DateConstRef AstronomicalMath::Date::operator+=(SFXDateDurationConstRef duration)
{
    return operator+=(DateDuration(duration));
}

inline AstronomicalMath::DateConstRef AstronomicalMath::Date::operator-=(DateDurationConstRef duration)
{
    _days -= duration.Get();
    _dateUpdated = false;
    _j2000Updated = false;
    return *this;
}

inline AstronomicalMath::DateConstRef AstronomicalMath::Date::operator-=(SFXDateDurationConstRef duration)
{
    return operator-=(DateDuration(duration));
}

inline Bool operator<(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2)
{
    return date2.operator>(date1);
}

inline AstronomicalMath::Date operator+(SFXDateDurationConstRef duration, AstronomicalMath::DateConstRef date)
{
    return date.operator+(duration);
}

inline Void AstronomicalMath::DateDuration::Set(SInt32 days, UInt16 hour, UInt16 minute, UInt16 second)
{
    _duration = static_cast<Float64>(days) + static_cast<Float64>(hour) / 24.0 + static_cast<Float64>(minute) / 1440.0 + static_cast<Float64>(second) / 86400.0;
}

inline Void AstronomicalMath::DateDuration::Set(Float64 days)
{
    _duration = days;
}

inline Void AstronomicalMath::DateDuration::Set(SFXDateDurationConstRef duration)
{
    _duration = static_cast<Float64>(duration.AsSInt64()) / 86400.0;
}

inline Void AstronomicalMath::DateDuration::Set(DateDurationConstRef ob)
{
    _duration = ob._duration;
}

inline AstronomicalMath::DateDurationConstRef AstronomicalMath::DateDuration::operator=(SFXDateDurationConstRef duration)
{
    Set(duration);
    return *this;
}

inline AstronomicalMath::DateDurationConstRef AstronomicalMath::DateDuration::operator=(DateDurationConstRef ob)
{
    Set(ob);
    return *this;
}

inline Float64 AstronomicalMath::DateDuration::Get(Void) const
{
    return _duration;
}

inline SInt32 AstronomicalMath::DateDuration::GetDays(Void) const
{
    return static_cast<SInt32>(floor(_duration));
}

inline UInt16 AstronomicalMath::DateDuration::GetHour(Void) const
{
    Float64Const hour((_duration - floor(_duration)) * 24.0);
    return static_cast<UInt16>(floor(hour));
}

inline UInt16 AstronomicalMath::DateDuration::GetMinute(Void) const
{
    Float64Const hour((_duration - floor(_duration)) * 24.0);
    Float64Const minute((hour - floor(hour)) * 60.0);
    return static_cast<UInt16>(floor(minute));
}

inline UInt16 AstronomicalMath::DateDuration::GetSecond(Void) const
{
    Float64Const hour((_duration - floor(_duration)) * 24.0);
    Float64Const minute((hour - floor(hour)) * 60.0);
    Float64Const second((minute - floor(minute)) * 60.0);
    return static_cast<UInt16>(floor(second + 0.5));
}

inline SFXDateDuration AstronomicalMath::DateDuration::GetSFXDateDuration(Void) const
{
    SInt32Const second(static_cast<SInt32>(floor(_duration * 86400.0 + 0.5)));
    return SFXDateDuration(second);
}

inline Bool AstronomicalMath::DateDuration::operator==(SFXDateDurationConstRef duration) const
{
    return operator==(DateDuration(duration));
}

inline Bool AstronomicalMath::DateDuration::operator==(DateDurationConstRef duration) const
{
    return _duration == duration._duration;
}

inline Bool operator==(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2)
{
    return duration2.operator==(duration1);
}

inline Bool AstronomicalMath::DateDuration::operator!=(SFXDateDurationConstRef duration) const
{
    return operator!=(DateDuration(duration));
}

inline Bool AstronomicalMath::DateDuration::operator!=(DateDurationConstRef duration) const
{
    return _duration != duration._duration;
}

inline Bool operator!=(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2)
{
    return duration2.operator!=(duration1);
}

inline Bool AstronomicalMath::DateDuration::operator>=(SFXDateDurationConstRef duration) const
{
    return operator>=(DateDuration(duration));
}

inline Bool AstronomicalMath::DateDuration::operator>=(DateDurationConstRef duration) const
{
    return _duration >= duration._duration;
}

inline Bool operator>=(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2)
{
    return duration2.operator<=(duration1);
}

inline Bool AstronomicalMath::DateDuration::operator>(SFXDateDurationConstRef duration) const
{
    return operator>(DateDuration(duration));
}

inline Bool AstronomicalMath::DateDuration::operator>(DateDurationConstRef duration) const
{
    return _duration > duration._duration;
}

inline Bool operator>(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2)
{
    return duration2.operator<(duration1);
}

inline Bool AstronomicalMath::DateDuration::operator<=(SFXDateDurationConstRef duration) const
{
    return operator<=(DateDuration(duration));
}

inline Bool AstronomicalMath::DateDuration::operator<=(DateDurationConstRef duration) const
{
    return _duration <= duration._duration;
}

inline Bool operator<=(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2)
{
    return duration2.operator>=(duration1);
}

inline Bool AstronomicalMath::DateDuration::operator<(SFXDateDurationConstRef duration) const
{
    return operator<(DateDuration(duration));
}

inline Bool AstronomicalMath::DateDuration::operator<(DateDurationConstRef duration) const
{
    return _duration < duration._duration;
}

inline Bool operator<(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2)
{
    return duration2.operator>(duration1);
}

inline AstronomicalMath::Date AstronomicalMath::DateDuration::operator+(DateConstRef date) const
{
    return Date(_duration + date.Get());
}

inline AstronomicalMath::Date AstronomicalMath::DateDuration::operator+(SFXDateConstRef date) const
{
    return operator+(Date(date));
}

inline AstronomicalMath::DateDuration AstronomicalMath::DateDuration::operator+(DateDurationConstRef duration) const
{
    return DateDuration(_duration + duration._duration);
}

inline AstronomicalMath::DateDuration AstronomicalMath::DateDuration::operator+(SFXDateDurationConstRef duration) const
{
    return operator+(DateDuration(duration));
}

inline AstronomicalMath::DateDuration AstronomicalMath::DateDuration::operator-(DateDurationConstRef duration) const
{
    return DateDuration(_duration - duration._duration);
}

inline AstronomicalMath::DateDuration AstronomicalMath::DateDuration::operator-(SFXDateDurationConstRef duration) const
{
    return operator-(DateDuration(duration));
}

inline AstronomicalMath::DateDurationConstRef AstronomicalMath::DateDuration::operator+=(DateDurationConstRef duration)
{
    _duration += duration._duration;
    return *this;
}

inline AstronomicalMath::DateDurationConstRef AstronomicalMath::DateDuration::operator+=(SFXDateDurationConstRef duration)
{
    return operator+=(DateDuration(duration));
}

inline AstronomicalMath::DateDurationConstRef AstronomicalMath::DateDuration::operator-=(DateDurationConstRef duration)
{
    _duration -= duration._duration;
    return *this;
}

inline AstronomicalMath::DateDurationConstRef AstronomicalMath::DateDuration::operator-=(SFXDateDurationConstRef duration)
{
    return operator-=(DateDuration(duration));
}

inline AstronomicalMath::DateDuration operator-(SFXDateConstRef date1, AstronomicalMath::DateConstRef date2)
{
    AstronomicalMath::Date t(date1);
    return t.operator-(date2);
}

inline AstronomicalMath::Date operator+(SFXDateConstRef date, AstronomicalMath::DateDurationConstRef duration)
{
    return duration.operator+(date);
}

inline AstronomicalMath::DateDuration operator+(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2)
{
    return duration2.operator+(duration1);
}

inline AstronomicalMath::DateDuration operator-(SFXDateDurationConstRef duration1, AstronomicalMath::DateDurationConstRef duration2)
{
    AstronomicalMath::DateDuration t(duration1);
    return t.operator-(duration2);
}

//
// The ecliptic inclination is calculated.
// T : Passage Julius year.
//
inline Float64 AstronomicalMath::CalcObliquityOfTheEcliptic(Float64 T)
{
    return 23.439291 - 0.000130042 * T;
}

//
// Check if it is an opposite sign or not (opposite sign = true).
//
inline Bool AstronomicalMath::DifferentSign(Float64 a, Float64 b)
{
    return (a < 0.0 && b > 0.0) || (a > 0.0 && b < 0.0);
}

//
// Get the date maintained by GMT.
//
inline AstronomicalMath::DateConstRef AstronomicalMath::Star::GetGMTDate(Void) const
{
    return _gmtDate;
}

//
// Get the ecliptic maintained coordinates.
//
inline AstronomicalMath::AstronomicalPointConstRef AstronomicalMath::Star::GetEclipticPosition(Void) const
{
    return _eclipticPosition;
}

//
// Get the maintained parallax.
//
inline Float64 AstronomicalMath::Star::GetParallax(Void) const
{
    return _PI;
}

//
// The celestial latitude of the sun is calculated (always 0).
// T : Passage Julius year.
//
inline Float64 AstronomicalMath::Sun::CalcEclipticLatitude(Float64 /* T */) const
{
    return 0.0;
}

//
// The parallax of the sun is calculated (not used for the sun calculations).
// T : Passage Julius year.
//
inline Float64 AstronomicalMath::Sun::CalcParallax(Float64 /* T*/) const
{
    return 0.0;
}

//
// Get the frequency of rise.
//
inline SInt32 AstronomicalMath::StarRiseSet::GetRiseDateSize(Void) const
{
    return _riseDate.GetSize();
}

//
// Get the date of rise.
//
inline AstronomicalMath::DateConstRef AstronomicalMath::StarRiseSet::GetRiseDate(SInt32 index) const
{
    return _riseDate.Get(index);
}

//
// Get the frequency of set.
//
inline SInt32 AstronomicalMath::StarRiseSet::GetSetDateSize(Void) const
{
    return _setDate.GetSize();
}

//
// Get the date of set.
//
inline AstronomicalMath::DateConstRef AstronomicalMath::StarRiseSet::GetSetDate(SInt32 index) const
{
    return _setDate.Get(index);
}

//
// Movement equation differentiation.
//
inline Float64 AstronomicalMath::StarRiseSet::DiffEquation(DateConstRef gmtDate) const
{
    DateDuration h(0, 0, 1, 0), h2(0, 0, 2, 0);

    // The movement equation argument is seen every day
    return 120.0 * (Equation(gmtDate - h2) - 8.0 * Equation(gmtDate - h) + 8.0 * Equation(gmtDate + h) - Equation(gmtDate + h2));
}

#endif
