//
//     Copyright (C) 2006 - 2007 SophiaCradle, Inc.
//     All rights reserved.
//
//      CityInformation.cpp
//

#include "CityInformation.hpp"

#define CITYINFORMATION_FILE_HEADER     "# City data\r\n# City name, Time difference (+0930 = +9Hour30Minute), Longitude, Latitude, Height (m)"

CityInformation::CityInformation(Void) : _changed(false)
{
    SFXFile file;
    SFXBinaryStreamReader reader;
    SFXAnsiString string;
    CITYINNER city;

    if (file.OpenReadOnly(SFXPath(CITYINFORMATION_FILENAME)) == SFERR_NO_ERROR) {
        if (file.GetStreamReader(1024, &reader) == SFERR_NO_ERROR) {
            while (!reader.Ends()) {
                ReadLine(&string, &reader);
                if (ParseText(&city, string)) {
                    _array.Append(city);
                }
            }
        }
        file.Close();
    }
    if (_array.GetSize() < 1) {
        // When the city information file is not found
        city._name.Set("Tokyo");
        city._timeDifference.Set(9 * 3600);
        city._longitude.Set("139.741");
        city._latitude.Set("35.658");
        city._height.Set("0.0");
        if (_array.Append(city) == SFERR_NO_ERROR) {
            _changed = true;
        }
    }
}

//
// Save to the file if there is a change.
//
CityInformation::~CityInformation(Void)
{
    SFXFile file;
    SFXAnsiStringStreamWriter writer;
    SFXAnsiString outString(CITYINFORMATION_FILE_HEADER), str;
    SInt32 i, val;
    CITYINNER city;

    if (_changed && _array.GetSize() > 0) {
        for (i = 0; i < _array.GetSize(); ++i) {
            outString.Add("\r\n");
            city = _array.Get(i);
            outString.Add(city._name);
            val = static_cast<SInt32>(city._timeDifference.AsSInt64());
            if (val > 0) {
                outString.Add(",+");
            }
            else {
                outString.Add(",-");
            }
            outString.Add(SFXAnsiString::Format("%02d%02d", val / 3600, (val % 3600) / 60));
            outString.Add(',');
            outString.Add(city._longitude);
            outString.Add(',');
            outString.Add(city._latitude);
            outString.Add(',');
            outString.Add(city._height);
        }

        SFXFile::Remove(SFXPath(CITYINFORMATION_FILENAME));
        if (file.OpenReadWrite(SFXPath(CITYINFORMATION_FILENAME)) == SFERR_NO_ERROR) {
            if (file.GetStreamWriter(outString.GetLength(), &writer) == SFERR_NO_ERROR) {
                if (writer.WriteSFXAnsiString(outString) == SFERR_NO_ERROR) {
                    writer.Flush();
                }
            }
            file.Close();
        }
    }
}

//
// The text file is read line by line (the null lines are skipped).
//
Void CityInformation::ReadLine(SFXAnsiStringPtr string, SFXBinaryStreamReaderPtr reader)
{
    UInt08 ch;
    Bool started(false);

    string->Clear();
    while (!reader->Ends()) {
        reader->Fetch();
        reader->ReadUInt08(&ch);
        if (ch == '\r' || ch == '\n') {
            if (!started) {
                // The line feed code at the head of the line is skipped
                continue;
            }
            else {
                // When the line feed code is found, it stops
                break;
            }
        }
        started = true;
        string->Add(ch);
    }
    return;
}

//
// A line of text is analyzed (the illegal content and the empty content return false).
//
Bool CityInformation::ParseText(CITYINNERPtr city, SFXAnsiStringConstRef string)
{
    SInt32 index, i, val;
    SFXAnsiString cleanedString, column[5];
    Bool result(true);

    // # is considered to be a comment
    index = string.FirstIndexOf('#');
    if (index == -1) {
        cleanedString = string;
    }
    else {
        cleanedString = string.Substring(0, index);
    }
    cleanedString.Trim();   // Extra blanks are removed

    // Each row is resolved
    // There are 5 rows
    // If extra data is added, disregard the error
    for (i = 0; i < 4; ++i) {
        index = cleanedString.FirstIndexOf(',');
        if (index == -1) {
            // The row doesn't suffice
            result = false;
            break;
        }
        column[i] = cleanedString.Substring(0, index);
        cleanedString = cleanedString.Remove(0, index + 1);
        cleanedString.Trim();
    }
    if (result) {
        // The error doesn't occur.
        if (cleanedString.IsEmptyCString()) {
            // The fifth row does not exist
            result = false;
        }
        if (column[0].IsEmptyCString()) {
            // No city name
            result = false;
        }
    }
    if (result) {
        index = cleanedString.FirstIndexOf(',');
        if (index == -1) {
            column[4] = cleanedString;
        }
        else {
            column[4] = cleanedString.Substring(0, index);
        }

        // The string is converted into the necessary form
        city->_name.Set(column[0]);
        val = column[1].AsSInt32();
        if (val < 0) {
            val = -val;
            val = - ((val / 100) * 3600 + (val % 100) * 60);
        }
        else {
            val = (val / 100) * 3600 + (val % 100) * 60;
        }
        city->_timeDifference.Set(val);
        city->_longitude.Set(column[2]);
        city->_latitude.Set(column[3]);
        city->_height.Set(column[4]);
    }

    return result;
}

//
// City information is acquired.
//
CityInformation::CITY CityInformation::Get(SInt32 index) const
{
    CITYINNERConst cityInner(_array.Get(index));
    CITY city;
    ACharPtr p;

    city._name.Set(cityInner._name);
    city._timeDifference.Set(cityInner._timeDifference);
    city._longitude = SFXHelper::strtod(cityInner._longitude.GetCString(), &p);
    city._latitude = SFXHelper::strtod(cityInner._latitude.GetCString(), &p);
    city._height = SFXHelper::strtod(cityInner._height.GetCString(), &p);

    return city;
}

//
// City information is registered.
// Overwrite the data.
//
SFCError CityInformation::Append(SFXWideStringConstRef cityName, SFXWideStringConstRef timeDifference, SFXWideStringConstRef longitude, SFXWideStringConstRef latitude, SFXWideStringConstRef height)
{
    CITYINNER city;
    SFCError result;
    SInt32 index, val;

    city._name.Set(cityName);
    val = timeDifference.AsSInt32();
    if (val < 0) {
        val = -val;
        val = - ((val / 100) * 3600 + (val % 100) * 60);
    }
    else {
        val = (val / 100) * 3600 + (val % 100) * 60;
    }
    city._timeDifference.Set(val);
    city._longitude.Set(longitude);
    city._latitude.Set(latitude);
    city._height.Set(height);

    if ((index = Find(cityName)) == -1) {
        // New registration
        result = _array.Append(city);
    }
    else {
        // Update data
        result = _array.Set(index, city);
    }

    if (result == SFERR_NO_ERROR) {
        _changed = true;
    }

    return result;
}

//
// City information is registered.
// Overwrite the data.
// appended : true = new registration.
// index : Registration position.
//
SFCError CityInformation::Append(SFXWideStringConstRef cityName, SFXWideStringConstRef timeDifference, SFXWideStringConstRef longitude, SFXWideStringConstRef latitude, SFXWideStringConstRef height, BoolPtr appended, SInt32Ptr index)
{
    CITYINNER city;
    SFCError result;
    SInt32 i, val;

    city._name.Set(cityName);
    val = timeDifference.AsSInt32();
    if (val < 0) {
        val = -val;
        val = - ((val / 100) * 3600 + (val % 100) * 60);
    }
    else {
        val = (val / 100) * 3600 + (val % 100) * 60;
    }
    city._timeDifference.Set(val);
    city._longitude.Set(longitude);
    city._latitude.Set(latitude);
    city._height.Set(height);

    if ((i = Find(cityName)) == -1) {
        // New registration
        result = _array.Append(city);
        if (result == SFERR_NO_ERROR) {
            *index = Find(cityName);
            *appended = true;
        }
        else {
            *index = -1;
            *appended = false;
        }
    }
    else {
        // Update data
        result = _array.Set(i, city);
        *index = i;
        *appended = false;
    }

    if (result == SFERR_NO_ERROR) {
        _changed = true;
    }

    return result;
}

//
// Look for city information by the name.
// If not existing returns -1.
//
SInt32 CityInformation::Find(SFXWideStringConstRef cityName) const
{
    CITYINNER city;
    SFXAnsiStringConst name(cityName);
    SInt32 i;

    for (i = 0; i < _array.GetSize(); ++i) {
        city = _array.Get(i);
        if (city._name == name) {
            return i;
        }
    }

    return -1;
}
